/*
master.c

Morgan Elder
CMPSCI 4760
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdbool.h>
#include "config.h"

// shared resources
typedef struct sharedData
{
  int *number;
  bool *choosing;
} shm_data;

shm_data *shared_mem;

key_t key = 5678;

int n;

static int isNumber(char *num_string);
static void doConcurrentProcs(int n, int timeout);
static int createSeg();
static int initSharedData(int id, int n);
static void detachSharedData();
static void removeSharedData(int id);

int main(int argc, char **argv)
{
  int opt;
  int timeout = MAX_TIMEOUT;
  n = MAX_PROCESSES;
  while ((opt = getopt(argc, argv, "t:")) != -1)
  {
    switch (opt)
    {
    case 't':
      if (optind != 3)
      {
        fprintf(stderr, "%s: error: Option %c must be the first argument\n", argv[0], opt);
        return 1;
      }
      else if (isNumber(optarg) == -1)
      {
        fprintf(stderr, "%s: error: arg for option %c must be integer value\n", argv[0], opt);
        return 1;
      }
      timeout = atoi(optarg);
      if (timeout > MAX_TIMEOUT)
      {
        fprintf(stderr, "%s: error: arg for option %c cannot exceed %d\n", argv[0], opt, MAX_TIMEOUT);
        return 1;
      }
      else if (timeout < MIN_TIMEOUT)
      {
        fprintf(stderr, "%s: error: arg for option %c cannot be negative\n", argv[0], opt);
        return 1;
      }
      break;
    case '?':
      fprintf(stderr, "%s: error: Uknown option %c\n", argv[0], optopt);
      return 1;
    case ':':
      fprintf(stderr, "%s: error: Missing arg for %c\n", argv[0], optopt);
      return 1;
    default:
      printf("Usage: %s -t ss n\n", argv[0]);
      return 0;
    }
  }

  if (optind > argc - 1)
  {
    fprintf(stderr, "%s: error: Missing arg specifying number of processes\n", argv[0]);
    return 1;
  }
  else if (isNumber(argv[argc - 1]) == -1)
  {
    fprintf(stderr, "%s: error: Arg specifying number of processes must be an number\n", argv[0]);
    return 1;
  }
  else
  {
    n = atoi(argv[argc - 1]);
    if (n > MAX_PROCESSES)
    {
      fprintf(stderr, "%s: error: Arg specifying number of processes cannot exceed %d\n", argv[0], MAX_PROCESSES);
      return 1;
    }
    else if (n < MIN_PROCESSES)
    {
      fprintf(stderr, "%s: error: Arg specifying number of processes cannot be less than %d\n", argv[0], MIN_PROCESSES);
      return 1;
    }
    doConcurrentProcs(n, timeout);
  }

  return 0;
}

static int isNumber(char *num_string)
{
  int i;
  for (i = num_string[0] == '-' ? 1 : 0; num_string[i] != '\0'; i++)
  {
    if (isdigit(num_string[i]))
    {
      continue;
    }
    else
    {
      return -1;
    }
  }
  return 0;
}

static void doConcurrentProcs(int n, int timeout)
{
  int shmid;
  int pid;

  // allocate shared memory
  // create segment with shmget(key, size, int shmflag)
  shmid = createSeg();

  // attach shared memory
  initSharedData(shmid, n);

  // only parent process can free shared memory resources
  // detach shared memory
  detachSharedData();

  // remove shared mem
  removeSharedData(shmid);
}

static int createSeg()
{
  printf("Enter createSeg()\n");
  int shmid;
  // key_t key = 5678;

  shmid = shmget(key, sizeof(shm_data), IPC_CREAT | 0600);
  if (shmid < 0)
  {
    perror("shmget");
    exit(EXIT_FAILURE);
  }
  return shmid;
}

static int initSharedData(int id, int n)
{

  // attach shared memory
  shared_mem = (shm_data *)shmat(id, NULL, 0);
  if (!shared_mem)
  {
    perror("shmat");
    exit(EXIT_FAILURE);
  }
  shared_mem->choosing = (bool *)malloc(n * sizeof(bool));
  if (shared_mem->choosing == NULL)
  {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  shared_mem->number = (int *)malloc(n * sizeof(int));
  if (shared_mem->number == NULL)
  {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  int i;
  for (i = 0; i < n; i++)
  {
    shared_mem->choosing[i] = false;
    shared_mem->number[i] = 0;
  }
  return 0;
}

static void detachSharedData()
{

  free(shared_mem->choosing);
  free(shared_mem->number);
  if (shmdt(shared_mem) == -1)
  {
    perror("shmdt");
    exit(EXIT_FAILURE);
  }
}

static void removeSharedData(int id)
{

  if (shmctl(id, IPC_RMID, NULL) == -1)
  {
    perror("shmctl");
    exit(EXIT_FAILURE);
  }
}